--*******************************************************************************************************************************
--
-- Description: This module performs all PCI Express write/read operations to base address register (BAR) zero.
--				The AXI_WIDTH constant determines whether slave_dataout is 64-bit or 128-bit aligned.
--					
-- Rev History: 09-04-2013:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.model_specific_package.all;
use work.sigfpga_package.all;

entity bar0_decoder is
port (
    -- Reset and clock.
    rst	                : in std_logic;									-- Active high asynchronous reset.
    clk                 : in std_logic;									-- Bus clock.

    -- PCI Express write transaction interface.
    slv_addr            : in std_logic_vector(13 downto 2);				-- Refer to the PLDA EZDMA reference manual.
    slv_dataout         : in std_logic_vector(AXI_WIDTH-1 downto 0);	-- Refer to the PLDA EZDMA reference manual.
    slv_bytevalid       : in std_logic_vector((AXI_WIDTH/8)-1 downto 0);-- Refer to the PLDA EZDMA reference manual.
    bar0_write          : in std_logic;									-- Logic one indicates an active slave write to BAR2 address space.

    -- PCI Express read transaction interface.
	bar0_read			: in std_logic;									-- Logic one indicates that the current PCI read request is directed at BAR0.
    bar0_rddata		    : out std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data return from read operation on BAR0.
    dma_rd              : in std_logic;									-- Refer to the PLDA EZDMA reference manual.
    dma_rdaddr          : in  std_logic_vector(13 downto 3);			-- Refer to the PLDA EZDMA reference manual.
	dma_rdchannel 		: in  std_logic;								-- Refer to the PLDA EZDMA reference manual.

    -- DMA command and status.
	dma_current_page	: in CH_TYPE_10;								-- Page number of the active DMA buffer.
	dma_current_burst	: in CH_TYPE_16;								-- Number of the active DMA burst within a page.
	dma_page_address	: out CH_TYPE_64;								-- Page address stored in the BAR0 memory.
	dma_descriptor		: out CH_TYPE_32;								-- Page descriptor stored in the BAR0 memory.
	dma_engine			: in CH_1TO4_3;									-- Status of DMA engine busy (2), go (1), and enable (0) bits.

    -- Consolidated interrupt mask and status.
	global_int_mask 	: out std_logic;								-- Global interrupt mask to enable/disable all interrupts.
	global_status		: in std_logic_vector(4 downto 0);				-- Consolidated interrupt status bus.  
	int_timeout			: out std_logic_vector(27 downto 0);			-- Maximum time to hold an interrupt request active.
	int_response		: in std_logic_vector(27 downto 0);				-- Time taken to service the last interrupt or timeout value.

	-- Serial peripheral interface bus.
	serial_config    	: out std_logic_vector(6 downto 0);				-- Serial bus command word.
	serial_payload	    : out std_logic_vector(7 downto 0);				-- Serial bus payload (instruction + data).
	serial_start		: out std_logic;								-- Logic one initiates a serial bus transaction.
	serial_busy			: in std_logic;									-- Logic one indicates that the process is active.
	serial_shift		: in std_logic;									-- Logic one indicates that the register needs to perform a serial shift.
	serial_read			: in std_logic;									-- Registered duplicate of sdo.      

	-- BPI configuration flash interface bus.
	bpi_config  	  	: out std_logic_vector(BPI_ADD+20 downto 0);	-- Flash address, write data and control signals.
	bpi_start			: out std_logic;								-- Logic one initiates a BPI bus transaction.
	bpi_busy			: in std_logic;									-- Logic one indicates that the process is active.
	bpi_rd_data			: in std_logic_vector(15 downto 0);				-- Read data bus.

	-- Command and status.
	clk_init_start		: out std_logic;								--  Hardware initialization trigger requested over the PCIe bus.
	clk_init_busy		: in std_logic;									--  Hardware initialization state machine status.
    software_reset		: out std_logic;								-- Active high reset tied to a register for software control.

    -- Error status and monitoring.
	error_monitor 		: in std_logic_vector(8 downto 0);				-- Bus used to monitor FIFO overflow or underlow errors.
	error_status 		: out std_logic_vector(8 downto 0);				-- Bus used to convey error status since last queried.
	error_int_mask 		: out std_logic_vector(8 downto 0);				-- Mask bits to suppress interrupts based on error status updates.
	bar0_address_error 	: out std_logic_vector(1 downto 0)				-- Bus used to convey a BAR0 PCI read or write error.
);
end bar0_decoder;

architecture structural of bar0_decoder is

	signal bytevalid				: std_logic_vector(1 downto 0);		-- Logic one indicates that all PCI byte valids are active.
	signal error_status_reg			: std_logic_vector(8 downto 0);		-- Bus used to monitor error status since last cleared.                                        	
	signal error_status_read		: std_logic_vector(8 downto 0);		-- Bus used to store error status before a clear so it can be read.                                        	
	signal error_int_mask_reg		: std_logic_vector(8 downto 0);		-- Mask bits to suppress interrupts based on fault status updates.
    signal serial_config_reg    	: std_logic_vector(6 downto 0);		-- Serial port configuration settings.
    signal serial_payload_reg    	: std_logic_vector(63 downto 0);	-- Serial port payload bits.
    signal bpi_config_reg    		: std_logic_vector(BPI_ADD+20 downto 0);	-- BPI port configuration settings.
    signal global_int_mask_reg 		: std_logic;						-- Global interrupt mask to enable/disable all interrupts.
    signal software_reset_reg    	: std_logic;						-- Register to hold a soft reset active for one clock period.
    signal int_timeout_reg    		: std_logic_vector(27 downto 0);	-- Maximum time to hold an interrupt request active.
	signal dma_bram_bytevalid		: DUAL_1SLV;						-- Logic one indicates that all PCI byte valids are active.
    signal dma_bram_data			: DUAL_32;							-- DMA memory write data bus with lane steering.
    signal dma_bram_select_reg    	: std_logic_vector(15 downto 0);	-- Index to DMA page addresses and descriptors.
    signal dma_bram_we    			: std_logic;						-- DMA page address and descriptor memory write qualifier.
    signal dma_bram_page_we    		: std_logic_vector(4 downto 1);		-- DMA page address memory write enable.
    signal dma_page_data			: CH_TYPE_64;						-- DMA page address memory data bus.
    signal dma_bram_desc_we    		: std_logic_vector(4 downto 1);		-- DMA descriptor memory write enable.
    signal dma_desc_data			: CH_TYPE_32;						-- DMA descriptor memory data bus.
	signal dma_current_page_local	: CH_1TO4_10;						-- Page number of the active DMA buffer.
	signal dma_current_burst_local	: CH_1TO4_16;						-- Number of the active DMA burst within a page.

begin

--***************************************************** WRITE TRANSACTIONS *****************************************************

	-- Create an active high enable for write operations to the upper 32 register bits.
	bytevalid(1) <= (slv_bytevalid(7) and slv_bytevalid(6) and slv_bytevalid(5) and slv_bytevalid(4)) when slv_addr(3) = '0'
		else (slv_bytevalid((AXI_WIDTH/8)-1) and slv_bytevalid((AXI_WIDTH/8)-2) and slv_bytevalid((AXI_WIDTH/8)-3) and slv_bytevalid((AXI_WIDTH/8)-4));
	-- Create an active high enable for write operations to the lower 32 register bits.
	bytevalid(0) <= slv_bytevalid(3) and slv_bytevalid(2) and slv_bytevalid(1) and slv_bytevalid(0) when slv_addr(3) = '0'
		else slv_bytevalid((AXI_WIDTH/8)-5) and slv_bytevalid((AXI_WIDTH/8)-6) and slv_bytevalid((AXI_WIDTH/8)-7) and slv_bytevalid((AXI_WIDTH/8)-8);
        
    process (clk,rst)
    begin
        if rst='1' then
            error_status_reg <= (others => '0');
            error_status_read <= (others => '0');
            error_int_mask_reg <= (others => '0');
            serial_payload_reg <= (others => '0');
            serial_config_reg <= (others => '0');
            bpi_config_reg <= (others => '0');
            global_int_mask_reg <= '0';
			software_reset_reg <= '0';
			int_timeout_reg <= (others => '1');
			dma_bram_select_reg <= (others => '0');
			clk_init_start <= '0';
			bar0_address_error(0) <= '0';
			serial_start <= '0';
			
        elsif rising_edge(clk) then
        
            -- Target write decoder that loads slave data into the selected register for a target write completion.
            if bar0_write = '1' then
                case "00" & slv_addr(13 downto 3) & "000" is

                	-- Software commanded reset equivalent to power on reset.
					when x"0000" =>
						software_reset_reg <= '1';  -- Will return to zero when the global reset is applied.
						
                	-- Clock initialization command.
                    when x"0008" =>
						if bytevalid(0) = '1' then
							-- Issue a single period enable strobe.
							clk_init_start <= '1';
						end if;
						
					-- Error status update command defined outside of this case statement.
                    when x"0010" =>
                    	null;
					
					-- Error interrupt mask register settings.
                    when x"0018" =>
						if bytevalid(0) = '1' then error_int_mask_reg <= slv_dataout((AXI_WIDTH-64)+8 downto (AXI_WIDTH-64)); end if;
						
					-- The serial bus payload register is defined outside of this case statement.
                    when x"0020" =>
                    	null;
					
					-- Serial bus configuration register.
                    when x"0030" =>
						if bytevalid(0) = '1' then serial_config_reg <= slv_dataout(6 downto 0); end if;

 					-- Serial bus transaction command.
                   when x"0038" =>
                        if bytevalid(0) = '1' then
							-- Issue a single period enable strobe.
							serial_start <= '1';
                        end if;

					-- BPI configuration register.
                    when x"0040" =>
						if bytevalid(1) = '1' then bpi_config_reg(BPI_ADD+20 downto BPI_ADD+5) <= slv_dataout(47 downto 32); end if;
						if bytevalid(0) = '1' then
							bpi_config_reg(BPI_ADD+4 downto BPI_ADD+3) <= slv_dataout(29 downto 28);
							bpi_config_reg(BPI_ADD+2 downto BPI_ADD) <= slv_dataout(26 downto 24);
							bpi_config_reg(BPI_ADD-1 downto 0) <= slv_dataout(BPI_ADD-1 downto 0);
						end if;

 					-- BPI bus transaction command.
                   when x"0048" =>
                        if bytevalid(0) = '1' then
							-- Issue a single period enable strobe.
							bpi_start <= '1';
                        end if;

                	--  x"0100" is a read only register.
					
					-- Global interrupt mask register.
					when x"0108" =>
						if bytevalid(0) = '1' then
							global_int_mask_reg <= slv_dataout(AXI_WIDTH-64);
						end if;
						
                	--  x"0110" Read only register.
                	
					-- Interrupt timeout value.
					when x"0118" =>
						if bytevalid(0) = '1' then
							int_timeout_reg <= slv_dataout((AXI_WIDTH-64)+27 downto (AXI_WIDTH-64));
						end if;
						
                	--  x"1F00" Read only register.
                	
                	--  x"1F08" Read only register.
                	
					-- Index into the DMA page address and descriptor memories.
					when x"1FF8" =>
						if bytevalid(0) = '1' then dma_bram_select_reg <= slv_dataout((AXI_WIDTH-64)+15 downto (AXI_WIDTH-64)); end if;
						
					-- Return an error if the address is not valid.
                    when others  =>
						if slv_addr(13) = '0' then bar0_address_error(0) <= '1'; end if;
						
                end case;                                                                                                          
			else
				clk_init_start <= '0';
				serial_start <= '0';
				bpi_start <= '0';
				bar0_address_error(0) <= '0';
            end if;
            
            -- Decode the write transaction to address x"0010" that will simultaneously capture and clear the fault status bits.
			if bar0_write = '1' and bytevalid(0) = '1' and ("00" & slv_addr(13 downto 3) & "000" = x"0010") then
				-- Capture the error status so that it can be read by software.
				error_status_read <= error_status_reg;
				-- Clear the sticky error status bits by updating to the instantaneous status.
				error_status_reg <= error_monitor;
			else
				-- Update the error status using sticky bits.
				error_status_reg <= error_status_reg or error_monitor;
			end if;
			
            -- Write to the serial payload register at address x"0020".
            if bar0_write = '1' and bytevalid(0) = '1' and ("00" & slv_addr(13 downto 3) & "000" = x"0020") then
            	serial_payload_reg(31 downto 0) <= slv_dataout(31 downto 0);
            elsif bar0_write = '1' and bytevalid(1) = '1' and ("00" & slv_addr(13 downto 3) & "000" = x"0020") then
            	serial_payload_reg(63 downto 32) <= slv_dataout(63 downto 32);
            elsif serial_shift = '1' then
            	-- The serial payload register is also a serial shift register.
				serial_payload_reg <= serial_payload_reg(62 downto 0) & serial_read;
			end if;
                          
        end if;
    end process;

    -- This process needs to be independent of the master reset so that software reset will eventually clear.
    process (clk)
    begin
        if rising_edge(clk) then
        	-- Register the soft reset so that it will be held active for one clock period.
			software_reset <= software_reset_reg;
        end if;
    end process;
   
--***************************************************** READ TRANSACTIONS ******************************************************

    process (clk,rst)
    begin
        if rst = '1' then
        
			bar0_address_error(1) <= '0';
			bar0_rddata <= (others => '0');
			
        elsif rising_edge(clk) then
  
            -- Target read decoder that loads selected register contents onto the DMA bus for a target read completion.
            if dma_rdchannel = '1' and dma_rd = '1' and bar0_read = '1' then
            
            	-- Set the default value to zero for cases that do not assign all of the bits.
				bar0_rddata <= (others => '0');
				
                case "00" & dma_rdaddr & "000" is
                
                	--  x"0000" is a write only register.
                	
                	-- Clock initialization status.
					when x"0008" =>
						bar0_rddata(AXI_WIDTH-64) <= clk_init_busy;
											
					-- Error status.
                    when x"0010" =>
						bar0_rddata(8 downto 0) <= error_status_read;
                 
					-- Error interrupt mask register settings.
                    when x"0018" =>
						bar0_rddata((AXI_WIDTH-64)+8 downto (AXI_WIDTH-64)) <= error_int_mask_reg;
                 
					-- Serial bus payload register.
                    when x"0020" =>
						bar0_rddata(63 downto 0) <= serial_payload_reg;

					-- Serial bus configuration register.
                    when x"0030" =>
						bar0_rddata(6 downto 0) <= serial_config_reg;

					-- Serial bus status.
                    when x"0038" =>
                        bar0_rddata(AXI_WIDTH-64) <= serial_busy;

					-- Serial bus configuration register.
                    when x"0040" =>
						bar0_rddata(63 downto 48) <= bpi_rd_data;
						bar0_rddata(47 downto 32) <= bpi_config_reg(BPI_ADD+20 downto BPI_ADD+5);
						bar0_rddata(29 downto 28) <= bpi_config_reg(BPI_ADD+4 downto BPI_ADD+3);
						bar0_rddata(26 downto 24) <= bpi_config_reg(BPI_ADD+2 downto BPI_ADD);
						bar0_rddata(BPI_ADD-1 downto 0) <= bpi_config_reg(BPI_ADD-1 downto 0);

					-- Serial bus status.
                    when x"0048" =>
                        bar0_rddata(AXI_WIDTH-64) <= bpi_busy;

					-- Global status.
					when x"0100" =>
						bar0_rddata(32) <= global_status(0);
						bar0_rddata(12) <= global_status(4);
						bar0_rddata(8) <= global_status(3);
						bar0_rddata(4) <= global_status(2);
						bar0_rddata(0) <= global_status(1);
						
					-- Global interrupt mask register.
					when x"0108" =>
						bar0_rddata(AXI_WIDTH-64) <= global_int_mask_reg;
						
					-- Measured interrupt response time.
					when x"0110" =>
						bar0_rddata(27 downto 0) <= int_response;
						
					-- Interrupt timeout value.
					when x"0118" =>
						bar0_rddata((AXI_WIDTH-64)+27 downto (AXI_WIDTH-64)) <= int_timeout_reg;
						
					-- DMA channel 1 and channel 2 status.
					when x"1F00" =>
						bar0_rddata(63 downto 48) <= dma_current_burst_local(2);
						bar0_rddata(46 downto 44) <= dma_engine(2);
						bar0_rddata(41 downto 32) <= dma_current_page_local(2);
						bar0_rddata(31 downto 16) <= dma_current_burst_local(1);
						bar0_rddata(14 downto 12) <= dma_engine(1);
						bar0_rddata(9 downto 0) <= dma_current_page_local(1);
						
					-- DMA channel 3 and channel 4 status.
					when x"1F08" =>
						bar0_rddata((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+48) <= dma_current_burst_local(4);
						bar0_rddata((AXI_WIDTH-64)+46 downto (AXI_WIDTH-64)+44) <= dma_engine(4);
						bar0_rddata((AXI_WIDTH-64)+41 downto(AXI_WIDTH-64)+32) <= dma_current_page_local(4);
						bar0_rddata((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64)+16) <= dma_current_burst_local(3);
						bar0_rddata((AXI_WIDTH-64)+14 downto (AXI_WIDTH-64)+12) <= dma_engine(3);
						bar0_rddata((AXI_WIDTH-64)+9 downto (AXI_WIDTH-64)) <= dma_current_page_local(3);
						
					-- Index into the DMA page address and descriptor memories.
					when x"1FF8" =>
						bar0_rddata((AXI_WIDTH-64)+15 downto (AXI_WIDTH-64)) <= dma_bram_select_reg;
						
					-- Return an error if the address is not valid.
                    when others  =>
						bar0_rddata(31 downto 0) <= x"badaddb0";
						if dma_rdaddr(13) = '0' then bar0_address_error(1) <= '1'; end if;
						
                end case;
			else
				bar0_address_error(1) <= '0';
            end if;
            
        end if;
    end process;
       
--***************************************************** SIGNAL ASSIGNMENTS *****************************************************

    error_status <= error_status_reg;
    error_int_mask <= error_int_mask_reg;
    global_int_mask <= global_int_mask_reg;
	int_timeout <= int_timeout_reg;
    serial_config <= serial_config_reg;
    serial_payload <= serial_payload_reg(63) & serial_payload_reg(55) & serial_payload_reg(47) & serial_payload_reg(39) & serial_payload_reg(31) & serial_payload_reg(23) & serial_payload_reg(15) & serial_payload_reg(7);
    bpi_config <= bpi_config_reg;

--************************************ DMA PAGE ADDRESS AND DESCRIPTOR MEMORY TRANSACTIONS *************************************

	-- Decode target write transactions to the DMA page address and descriptor memories.
	dma_bram_we <= bar0_write and slv_addr(13);
	
	-- Convert the byte valid flag bits to standard logic vector so that they match the BRAM port type.
	dma_bram_bytevalid(0)(0) <= bytevalid(0);
	dma_bram_bytevalid(1)(0) <= bytevalid(1);
	
	-- Lane steering depends on the width of the AXI bus.
	dma_bram_data(1) <= slv_dataout(63 downto 32) when slv_addr(3) = '0' else slv_dataout((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+32);
	dma_bram_data(0) <= slv_dataout(31 downto 0) when slv_addr(3) = '0' else slv_dataout((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64));
	
	-- Generate a 64-bit page address memory for each DMA channel.
	PAGE_MEMORY : for i in 1 to TX_CHANNEL_COUNT+RX_CHANNEL_COUNT generate
		dma_bram_page_we(i) <= dma_bram_we and dma_bram_select_reg(i);
			-- Use two 32-bit BRAMs to emulate a register write operation.
			BRAM_X2 : for j in 0 to 1 generate
				PAGE_BRAM : entity work.dma_on_demand_memory
					port map(
						clka => clk,
						ena => dma_bram_page_we(i),
						wea => dma_bram_bytevalid(j),
						addra => slv_addr(12 downto 3),
						dina => dma_bram_data(j),
						clkb => clk,
						addrb => dma_current_page(i),
						doutb => dma_page_data(i)((32*j)+31 downto 32*j)
						);
			end generate;
		dma_page_address(i) <= dma_page_data(i)(63 downto 32) & dma_page_data(i)(31 downto 0);
	end generate;
			
	-- Generate a 32-bit descriptor memory for each DMA channel.
	DESC_MEMORY : for i in 1 to TX_CHANNEL_COUNT+RX_CHANNEL_COUNT generate
		dma_bram_desc_we(i) <= dma_bram_we and dma_bram_select_reg(i+8);
		DESC_BRAM : entity work.dma_on_demand_memory
			port map(
				clka => clk,
				ena => dma_bram_desc_we(i),
				wea => dma_bram_bytevalid(0),
				addra => slv_addr(12 downto 3),
				dina => dma_bram_data(0),
				clkb => clk,
				addrb => dma_current_page(i),
				doutb => dma_desc_data(i)
				);
		dma_descriptor(i) <= dma_desc_data(i);
	end generate;

	-- Make signal assignments to unused DMA channels.
	DMA_CH : for i in 1 to 4 generate
		CH_ACTIVE: if i < TX_CHANNEL_COUNT+RX_CHANNEL_COUNT+1 generate
			dma_current_page_local(i) <= dma_current_page(i);
			dma_current_burst_local(i) <= dma_current_burst(i);
		end generate;	
		CH_UNUSED: if i > TX_CHANNEL_COUNT+RX_CHANNEL_COUNT generate
			dma_current_page_local(i) <= (others=>'0');
			dma_current_burst_local(i) <= (others=>'0');
		end generate;	
	end generate;	
		
end structural;
