--*******************************************************************************************************************************
--
-- Description: This module is used to dynamically optimize the FPGA IOB timing at the chip interface so that static timing
--				constraints are not necessary.  This procedure only needs to be run once, typically when the hardware
--				is initialized following power-up.  A state machine adjusts the phase relationship of incoming data to 
--				the global clock registering the data so that setup and hold margins are maximized.
--					
--				This core is designed to lock clocks from multiple sources, ensuring that all sources are synchronized
--				to the same edge of global clock.
--					
-- Rev History: 8-21-2012:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all; 

Library UNISIM; 
use UNISIM.Vcomponents.all; 

entity clock_lock is

generic(
	SOURCES : integer;
	MAX_SEARCH : integer;
	MAX_SEPARATION : integer;
	RESET_IF_NOSOL : boolean;
	FINAL_OFFSET : integer
);

port (
--*******************************************************************************************************************************
-- The following user interface signals should be connected to application logic internal to the FPGA.

	-- Connect this input to the global reset in the application logic.
	rst         	: in std_logic;		-- Active high asynchronous reset.       

	-- Signals used to trigger and monitor the clock lock function.
	lock_start		: in std_logic; 	-- Single period enable initiates the clock lock sequence that sets IODELAY values for data inputs. 
	lock_busy		: out std_logic; 	-- Logic one indicates that the clock lock process is active. 
	lock_done		: out std_logic; 	-- Single period lock done pulse to indicate that the process is complete. 
	lock_status		: out std_logic;	-- Logic one indicates that a solution was found.
										-- Logic zero does not necessarily indicate a failure to lock.
		
	-- Connections to the IODELAY function in the IOB that needs to be phase aligned.
	clk_lck_clk		: in std_logic;								-- Reference clock used to optimize IOB timing.
	clk_fwd			: in std_logic_vector(SOURCES-1 downto 0);	-- Forwarded clock signal registered in the IOB.
	clk_lck_edge	: in std_logic_vector(1 downto 0);			-- Command to the clock lock function to look for specific edges.
	idly_rst		: out std_logic;							-- Logic one enables the IODELAY reset function.
	idly_en			: out std_logic_vector(SOURCES-1 downto 0);	-- Logic one enables the IODELAY increment/decrement function.
	inc_dec			: out std_logic_vector(SOURCES-1 downto 0)	-- Logic one increments the IODELAY one tap and logic zero decrements one tap.
);
end clock_lock;

architecture behavioral of clock_lock is

	type FIR is array (SOURCES-1 downto 0) of unsigned(8 downto 0);		-- Type definition for a small 1-bit FIR filter.
	signal tap				: FIR;										-- Nine tap delay line to filter the forwarded clock.
	type FILTER is array (SOURCES-1 downto 0) of unsigned(2 downto 0);	-- Type definition for filter results.
	signal filter_lead		: FILTER;									-- Leading edge filter calculation result.
	signal filter_trail		: FILTER;									-- Trailing edge filter calculation result.
		
	type CLK_LCK_STATE is (idle, reset, init, search, gather, nosol, offset, done);	-- Type definitions for the state machine.
	signal idly_state 		: CLK_LCK_STATE;				-- Current state.
	signal idly_en_local	: unsigned(SOURCES-1 downto 0);	-- Local IODELAY enable flags.
	signal idly_counter		: unsigned(5 downto 0);			-- IODELAY increment counter.
	signal start_search		: unsigned(5 downto 0);			-- Starting count for the IODELAY search function.
	signal reset_counter	: unsigned(3 downto 0);			-- Counter to stretch the IODELAY reset.
	signal no_solution		: std_logic; 					-- Active high flag to indicate that the search did not find a solution.
	
begin                    

--************************************************* STATE MACHINE TRANSITIONS **************************************************

	process(rst, clk_lck_clk)
	begin
		if rst = '1' then
			idly_state <= idle;
			
		elsif rising_edge(clk_lck_clk) then
			
			-- The following state machine searches for the IODELAY setting that maximizes timing margin.
			case idly_state is			
				-- Remain idle while waiting for the application logic to initiate a clock lock sequence.
				when idle =>
					-- Detect the lock_start pulse to begin the clock lock process.	
					if lock_start = '1' then
						idly_state <= reset;
					end if;
				-- Reset the control logic.
				when reset =>
					-- Hold the reset for sixteen clock periods to stabilize the initial delay values and clear the filter.	
					if reset_counter = x"F" and no_solution = '1' then
						idly_state <= done;
					elsif reset_counter = x"F" and no_solution = '0' then
						idly_state <= init;
					end if;
				-- Initialize the search criteria.
				when init =>
					if idly_counter = start_search then
						idly_state <= search;
					end if;
				-- Find the requested edge of the forwarded clocks.
				when search =>
					-- The search is complete when an edge is detected.
					if to_integer(idly_en_local) /= (2**SOURCES)-1 then	
						idly_state <= gather;
					-- The search is terminated if the maximum IODELAY count is reached.
					elsif idly_counter = to_unsigned(MAX_SEARCH,6) then
						idly_state <= nosol;
					end if;
				-- Find the requested edge of all remaining clocks within the maximum specified separation.
				when gather =>
					-- The gather is complete when all of the clock edges are detected.
					if to_integer(idly_en_local) = 0 then	
						idly_state <= offset;
					-- The gather is terminated if the maximum IODELAY count is reached.
					elsif idly_counter = to_unsigned(MAX_SEARCH,6) then
						idly_state <= nosol;
					-- The search begins again if the count has exceeded the maximum separation between clocks.
					elsif idly_counter = (start_search + to_unsigned(MAX_SEPARATION,6)) then
						idly_state <= reset;
					end if;
				-- Clear the ready flag if the maximum IODELAY count is reached before a solution is found.
				when nosol =>
					if RESET_IF_NOSOL = true then
						idly_state <= reset;
					else
						idly_state <= done;
					end if;
				-- Increment the requested offset from the computed solution.
				when offset =>
					if idly_counter = to_unsigned(FINAL_OFFSET,6) then
						idly_state <= done;
					end if;
				-- Generate a single clock period done pulse.
				when done =>
					idly_state <= idle;
				-- Define the default state to be idle.
				when others  =>
						idly_state <= idle;
			end case;
			
		end if;
	end process;
	
--*************************************************** STATE MACHINE EVENTS *****************************************************
			
	process(rst, clk_lck_clk)
	begin
		if rst = '1' then
			lock_status <= '0';
			for i in 0 to SOURCES-1 loop
				tap(i) <= (others => '0');
				filter_lead(i) <= (others => '0');
				filter_trail(i) <= (others => '0');
			end loop;
			reset_counter <= (others => '0');
			idly_rst <= '0';
			idly_counter <= (others => '0');
			idly_counter <= (others => '0');
			idly_en_local <= (others => '0');
			start_search <= (others => '0');
			no_solution <= '0';
			
		elsif rising_edge(clk_lck_clk) then
			
			-- Create ten tap delay lines for each source to use as a filter on the forwarded clock signal.
			for i in 0 to SOURCES-1 loop
				tap(i)(0) <= clk_fwd(i);
				for j in 0 to 7 loop
					tap(i)(j+1) <= tap(i)(j);
				end loop;
			end loop;
			
			-- Filter the leading and trailing edge of the clock transition on each input to eliminate false triggers.
			for i in 0 to SOURCES-1 loop
				filter_lead(i) <= ("00" & clk_fwd(i)) + ("00" & tap(i)(0)) + ("00" & tap(i)(1)) + ("00" & tap(i)(2)) + ("00" & tap(i)(3));
				filter_trail(i) <= ("00" & tap(i)(4)) + ("00" & tap(i)(5)) + ("00" & tap(i)(6)) + ("00" & tap(i)(7)) + ("00" & tap(i)(8));
			end loop;
			
			-- Enable and disable reset to the IODELAY counter in the IOB (idly_rst).
			if idly_state = reset then
				-- A counter is used to hold reset long enough to clear the clock edge filter.
				reset_counter <= reset_counter + x"1";
				-- The IOB is reset to the initial IODELAY value.
				if reset_counter < "0100" then
					idly_rst <= '1';
				else
					idly_rst <= '0';
				end if;
			else
				-- Hold the reset inactive.
				reset_counter <= (others => '0');
				idly_rst <= '0';
			end if;
					
			-- Set the enable bit on the IOB IODELAY counter (idly_en_local) and track the count (idly_counter).
			if idly_state = init then
				-- The IODELAY counter must be initialized to the starting value (start_search) for multi-search cases.
				-- Note: The filter is simply primed with one sample for single search cases, so there is no loss of range.
				idly_counter <= idly_counter + "000001";
				-- Enable the IODELAY values to increment.
				idly_en_local <= (others => '1');
			elsif (idly_state = search) or (idly_state = gather) then
				-- The IODELAY counter increments on every clock to track the tap value in the IOB.
				if to_integer(idly_en_local) = 0 then
					-- Reset the counter if a solution is found to prepare for the offset adjustment.
					idly_counter <= (others => '0');
				else
					-- Increment the counter.
					idly_counter <= idly_counter + "000001";
				end if;
				-- Search for edges on any clock that hasn't been already been detected.
				for i in 0 to SOURCES-1 loop
					-- Detect the rising edge of the forwarded clocks while ignoring glitches one or two clocks in duration.
					if (clk_lck_edge(0) = '1') and (filter_lead(i) > "010") and (filter_trail(i) < "011") then
						idly_en_local(i) <= '0';
					-- Detect the falling edge of the forwarded clocks while ignoring glitches one or two clocks in duration.
					elsif (clk_lck_edge(1) = '1') and (filter_lead(i) < "011") and (filter_trail(i) > "010") then
						idly_en_local(i) <= '0';
					end if;
				end loop;
			elsif idly_state = offset then
				-- Add the requested offset to the computed solution.
				idly_counter <= idly_counter + "000001";
				-- Enable the IODELAY values to increment.
				idly_en_local <= (others => '1');
			else
				-- The IODELAY counter is cleared.
				idly_counter <= (others => '0');
				idly_en_local <= (others => '0');
			end if;

			-- Store the count value to start the next search (start_search).
			if idly_state = idle then
				-- Clear the start count.
				start_search <= (others => '0');
			elsif idly_state = search then
				-- Let the start count track the IODELAY count.
				start_search <= idly_counter;
			end if;
			
			-- Set the solution status flags to indicate whether a solution was found.
			if idly_state = idle then
				no_solution <= '0';
			elsif idly_state = reset then
				lock_status <= '1';
			elsif idly_state = nosol then
				lock_status <= '0';
				no_solution <= '1';
			end if;
			
		end if;
	end process;
	
--************************************************ INTERNAL SIGNAL ASSIGNMENTS *************************************************
			
	lock_busy <= '0' when idly_state = idle else '1';
	lock_done <= '1' when idly_state = done else '0';
	idly_en <= std_logic_vector(idly_en_local);
	inc_dec <= (others => '1');

end behavioral;
